import pytest
from odoo_tools.modules.search import Manifest
from odoo_tools.compat import Path
from odoo_tools.modules.search import get_manifest, find_modules_paths
from odoo_tools.api.objects import try_compile_manifest
from odoo_tools.exceptions import ArgumentError
from six import ensure_binary, ensure_text

import mock
from io import BytesIO, StringIO


fake_manifest_1 = """# coding: utf-8
# SOme comment
{
    "name": "Some",
    "depends": ["web", "base"],
    "version": "0.1",
}
"""

fake_manifest_2 = ensure_binary(fake_manifest_1)

no_name_manifest = """
{
  "depends": ["sale"],
}
"""


def test_manifest():
    manifest = Manifest('test', {})

    assert manifest.installable is True
    assert manifest.application is False
    assert manifest.depends == []
    assert manifest.version == "0.0.0"
    assert manifest.external_dependencies == {}
    assert manifest.path == Path("test")


def test_manifest_with_data():
    data = {
        "name": "Fun",
        "description": "Description",
        "installable": False,
        "application": True,
        "version": "0.1",
        "external_dependencies": {
            "python": ["gitpython"]
        },
        "depends": ["web", "sale"]
    }

    manifest = Manifest("test2", data)

    assert manifest.depends == ["web", "sale"]
    assert manifest.external_dependencies == {"python": ["gitpython"]}
    assert manifest.version == "0.1"
    assert manifest.installable is False
    assert manifest.application is True
    assert manifest.name == "Fun"


def test_read_string_manifest():
    """ test manifest read from string """
    fake_manifest = StringIO(ensure_text(fake_manifest_1))
    with mock.patch('odoo_tools.compat.Path.exists', return_value=True):
        with mock.patch(
            'odoo_tools.compat.Path.open',
            return_value=fake_manifest
        ):
            manifest = get_manifest(Path("my_modules/__manifest__.py"))

            assert manifest.name == "Some"
            assert manifest.technical_name == "my_modules"
            assert manifest.version == "0.1"


def test_read_bytes_manifest():
    """ Test a manifest read from bytes """
    fake_manifest = BytesIO(ensure_binary(fake_manifest_2))
    with mock.patch('odoo_tools.compat.Path.exists', return_value=True):
        with mock.patch(
            'odoo_tools.compat.Path.open',
            return_value=fake_manifest
        ):
            manifest = get_manifest(Path("my_modules/__manifest__.py"))

            assert manifest.name == "Some"
            assert manifest.technical_name == "my_modules"
            assert manifest.version == "0.1"


def test_noname():
    """Test a manifest without name or version"""
    fake_manifest = StringIO(ensure_text(no_name_manifest))

    with mock.patch('odoo_tools.compat.Path.exists', return_value=True):
        with mock.patch(
            'odoo_tools.compat.Path.open',
            return_value=fake_manifest
        ):
            manifest = get_manifest(Path("my_modules/__manifest__.py"))

            assert manifest.name == "my_modules"
            assert manifest.technical_name == "my_modules"
            assert manifest.depends == ['sale']
            assert manifest.version == "0.0.0"


def test_find_modules(tmp_path):
    paths = [
        "a/b/__init__.py",
        "a/b/__manifest__.py",
        "a/c/__init__.py",
        "a/c/__openerp__.py",
        ".git/a/b/d",
        ".git/a/e/d",
        ".git/a/e/f",
        ".git/a/e/g",
        "a/d/e/__init__.py",
        "a/d/e/__manifest__.py"
    ]

    for path in paths:
        pp = tmp_path / Path(path)

        if pp.name.endswith('.py'):
            pp.parent.mkdir(parents=True, exist_ok=True)
            pp.open('w').write(ensure_text(fake_manifest_1))
        else:
            pp.mkdir(parents=True, exist_ok=True)

    modules = find_modules_paths([tmp_path])
    assert len(modules) == 3


def test_manifest_operators(tmp_path):
    man1 = Manifest(tmp_path / 'a')
    man2 = Manifest(tmp_path / 'a')
    man3 = Manifest(tmp_path / 'b')
    man4 = Manifest(tmp_path / 'b' / 'a')

    assert (man1 == man2) is True
    assert (man1 < man3) is True
    assert (man1 < man4) is False
    assert (man1 == 'a') is True
    assert (man3 == 'b') is True
    assert (man1 == 0) is False
    assert str(man1) == str(tmp_path / 'a')
    assert repr(man1).startswith('Manifest(') is True

    vals = {
        "technical_name": "a",
        "depends": [],
        "application": False,
        "installable": True,
        "external_dependencies": {},
        "demo": [],
        "data": [],
        "version": "0.0.0",
    }

    assert man1.values() == vals

    man1.version = "1.0.0"
    vals['version'] = "1.0.0"

    assert man1.values() == vals
    man1.set_attribute(
        ['external_dependencies', 'python'],
        ['cryptography']
    )
    man1.set_attribute(
        ['external_dependencies', 'binary', 'test'],
        ['cryptography']
    )

    with pytest.raises(ArgumentError):
        man1.set_attribute([], ['cryptography'])

    vals['external_dependencies'] = {
        "python": ['cryptography'],
        "binary": {
            "test": ['cryptography']
        }
    }

    assert man1.values() == vals
    assert 'external_dependencies' in man1
    assert 'vacuum' not in man1


def test_try_eval_manifest():
    data = try_compile_manifest('{"name": "test"}')
    assert data == {"name": "test"}



"""
Search
======

The module search regroup a series of functions used to search
modules and handle their dependencies. This collection of functions,
can be used to determine which folder contains installable modules
and it can also be used to predetermine which module would require
to be installed if a given module was installed.
"""
import logging

from odoo_tools.compat import Path, module_path
from ..api.objects import Manifest


import pkg_resources

_logger = logging.getLogger(__name__)


def filter_installable(manifest):
    """
    Filter for installable modules
    """
    return manifest.installable


def filter_noninstallable(manifest):
    """
    Filter for not installable modules.

    This is mainly the opposite of `filter_installable`.
    """
    return not filter_installable(manifest)


def filter_python_dependencies(manifest):
    """
    Filter modules with python dependencies.
    """
    if (
        manifest.external_dependencies and
        manifest.external_dependencies.get('python')
    ):
        return True
    else:
        return False


def get_filter(filter_names):
    """
    Returns a filter function with the provided filter names.

    If all filter function returns True, then the object will
    not get filtered out of the set/list.

    For example, a filter of installable, python_dependencies on
    a list of manifests will return all manifests that are
    installable and that have python dependencies.

    Args:
        filter_names (Set(str)): Set of filter names.

    Returns:
        Boolean: if the manifest validates the current set of
            filters.
    """
    if not filter_names:
        return lambda module: True

    filters = []

    if 'installable' in filter_names:
        filters.append(filter_installable)

    if 'non_installable' in filter_names:
        filters.append(filter_noninstallable)

    if 'python_dependencies' in filter_names:
        filters.append(filter_python_dependencies)

    def filter_module(manifest):
        return all([check(manifest) for check in filters])

    return filter_module


def fast_search_manifests(path):
    """
    Quickly search into directoy for manifest files.

    In order to speed up the search recursively, it
    will stop search in folders that have one of the
    possible manifest name.

    It doesn't loop through all files in the directory.
    Instead it checks for manifest files then if none
    are found, it will lookup for folders to look into.

    All other files are skipped. Then the cycle repeats,
    until it finds a manifest or there are no more folders
    to search into.

    Args:
        path (Path): Path in which the manifest lookup occurs.

    Returns:
        list(Path): List of manifests paths.
    """
    filenames = ['__manifest__.py', '__openerp__.py']
    found_paths = []
    blacklist = ['setup', '.git']

    for manifest in filenames:
        manifest_path = path / manifest
        if manifest_path.exists():
            return [manifest_path]

    dirs_to_search = []
    if path.exists():
        for cpath in path.iterdir():
            if cpath.name in blacklist:
                continue
            if cpath.is_dir():
                dirs_to_search.append(cpath)
                continue
            # Never used because tested before looking up into dir
            # if cpath.name in filenames:
            #     found_paths.append(cpath)
            #     return
        else:
            for cpath in dirs_to_search:
                found_paths += fast_search_manifests(cpath)

    return found_paths


def find_modules(path, filters=None):
    """
    Search for manifests recursively in a specified folder.

    Each Path is then converted to a Manifest object.

    Args:
        path (Path): Path in which the manifest lookup occurs.

        filters (Set(str)): Set of filters to ignore some manifests.

    Returns:
        list(Manifest): A list of valid manifests.
    """
    modules = set()

    path = Path.cwd() / path

    manifest_globs = fast_search_manifests(path)

    check_module = get_filter(filters)

    for path in manifest_globs:
        manifest = Manifest.from_path(path)

        if not check_module(manifest):
            continue

        modules.add(manifest)

    return modules


def find_modules_paths(paths, filters=None, options=None):
    """
    Search modules in multiple paths.


    This can be used to search for odoo modules in different
    addons paths.

    For example, you'd want to discover all odoo modules installed
    in your environment.

    Args:
        paths (list(Path)): list of Path in which the manifest
            lookup occurs.

        filters (Set(str)): Set of filters to ignore some manifests.

        options (object): Object with a flag to exclude core odoo addons.

    Returns:
        list(Manifest): All manifests in all the paths provided.
    """
    modules = set()

    if options and not options.exclude_odoo:
        odoo_path = base_addons_path()
        if odoo_path:
            paths.add(odoo_path)

    for path in paths:
        modules = modules.union(
            find_modules(Path(path), filters=filters)
        )

    return modules


def get_manifest(manifest, render_description=False):
    """
    Shortcut to get a manifest from path.

    # TODO maybe this should be deprecated now.
    """
    return Manifest.from_path(manifest, render_description)


def build_module_dependencies(
    modules,
    modules_lst=False,
    deps=None,
    quiet=True
):
    """
    Generate dependencies of each modules passed in this function.

    When all modules data is loaded, it is then possible to generate
    a graph of all the modules and their dependencies.

    This graph then can be used in a topological sort to guess in which
    correct order the modules should be installed. Based on the
    dependencies, it's also possible to guess which modules are
    required by one other module and find which modules are missing.

    Args:
        modules (list(Manifest)): List of manifest to use for dependencies

        modules_lst (list(Manifest)): List of manifest to build dependencies
            for.

        deps (dict): Initial dependencies

        quiet (bool): Silence missing modules when they're not present in the
            module set.

    Returns:
        dict: A dictionary of {module: [dep,..], ...} of the current set.
    """
    if deps is None:
        deps = {}

    to_process = modules_lst or []

    while len(to_process) > 0:
        cur_module = to_process.pop()
        if cur_module not in modules:
            continue

        dependencies = modules[cur_module].depends
        deps[cur_module] = set(dependencies)

        for dep in dependencies:
            if dep not in deps:
                to_process.append(dep)

            if not quiet and dep not in modules:
                print((
                    "Module {cur_module} depends on {dep} "
                    "which isn't in addons_path"
                ).format(cur_module=cur_module, dep=dep))

    return deps


def build_dependencies(
    modules,
    modules_lst,
    lookup_auto_install=True,
    deps=None,
    quiet=True
):
    """
    Build dependencies with auto installable modules.

    It's more or less the same as build_module_dependencies, but
    it goes a step further by injecting auto dependent modules
    that have all their dependencies in the dependencies looked up
    first.

    This by itself makes it possible to guess the modules that would
    need to be installed given the current addons path and specified
    modules that needs to be checked.

    For example, if you wanted to check for the modules_lst = ['sale', 'stock']

    It would first pull all the dependencies of the module sale, then it would
    pull all the dependencies of the module stock. Then it would lookup
    for all modules that are auto installable. If all of their dependencies are
    in the pulled dependencies. They'd get pulled into the dependencies.

    Then it would check again for auto installable modules if any new
    dependency can pull more auto installable modules.

    Unfortunately, it's not possible to have 100% guarantee that you'll
    get the exact same module set as in odoo. There is a guarantee that all
    the modules found will be installed. But odoo can request more modules
    to be installed as the account module that trigger some extra modules
    to be installed at init time.

    Args:
        modules (list(Manifest)): List of all modules available

        modules_lst (list(Manifest)): List of manifest you want to find their
            dependencies
        lookup_auto_install (bool): Lookup for auto installable modules.

        deps (dict): List of already known dependencies

        quiet (bool): Silence missing dependencies logs

    Returns:
        dict: dict of dependencies of the format {module: [dep1,...], ...}
    """
    deps = build_module_dependencies(
        modules,
        modules_lst,
        deps=deps,
        quiet=quiet
    )

    if not lookup_auto_install:
        return deps

    old_deps_length = 0

    while len(deps) != old_deps_length:
        old_deps_length = len(deps)

        to_install = []

        for name, module in modules.items():
            if not module.auto_install or name in deps:
                continue

            module_deps = module.depends
            for dep in module_deps:
                if dep not in deps:
                    break
            else:
                if len(module_deps) > 0:
                    to_install.append(name)

        deps = build_module_dependencies(
            modules,
            to_install,
            deps=deps,
            quiet=quiet
        )

    return deps


def base_addons_path():
    """
    Find the path of odoo itself.

    Returns:
        Path: Location where odoo is installed.
    """
    odoo_path = (
        module_path("odoo", raise_not_found=False) or
        module_path("openerp", raise_not_found=False)
    )
    return odoo_path


def find_addons_paths(paths, options=False):
    """
    Find addons in provided paths.

    Args:
      paths (List<Path>): A list of paths to search for addons.

    Returns:
      Set<Manifest>: A list of manifests
    """
    filters = set()
    filters.add('installable')

    if options and not options.exclude_odoo:
        odoo_path = base_addons_path()
        if odoo_path:
            paths.add(odoo_path)

    if options and options.include_odoo_entrypoints:
        entry_point = "odoo_addons_paths"
        for ep in pkg_resources.iter_entry_points(group=entry_point):
            functor = ep.load()
            new_paths = functor()
            paths |= new_paths

    modules = find_modules_paths(
        paths,
        filters=filters
    )

    found_paths = set()

    for manifest in modules:
        found_paths.add(manifest.path.parent)

    return found_paths



import os
from ..compat import Path
from ..env import EnvironmentVariables


class Context(object):
    """
    The context in which you use the Environment.

    For example, the context will be able to load environment
    variables configured in the running environment.

    The environment variables are then passed to the context
    and then the application is able to have a proper behaviour
    based on the environment.

    Attributes:

        strict_mode (bool): Can be used to run in strict mode. In strict mode,
            the library doesn't ignore exceptions and raise exceptions to the
            caller.

            This is useful in CI pipelines that may want to know in case of
            failure. One example is when running Pip install modules, it may
            complete installing but should stop in case some dependencies
            can't be installed properly.

            In some cases, when running a docker image, you may want your
            docker container to keep running even if some pip dependencies
            can't be installed/upgraded.

        custom_paths (Set<Path>): A list of ``Path`` that contains modules.
            The path may not directly contain modules as this library can be
            used to find modules recursively in given paths.
    """
    def __init__(
        self,
        custom_paths=None,
        strict_mode=True,
        requirements_file_path=None,
        odoo_base_path=None,
        excluded_paths=None,
        exclude_odoo=False,
        disabled_modules=False,
        odoo_rc=False,
        include_odoo_entrypoints=True,
        force_addons_lookup=False,
        init_logger=True,
        run_only=False,
        extra_apt_packages=None,
        apt_install_recommends=False,
        package_map_file=None,
        skip_pip=False,
        skip_sudo_entrypoint=False,
        skip_postgres_wait=False,
        allow_dangerous_settings=False,
        odoo_version=None,
        master_password=None,
        show_master_password=True,
        reset_access_rights=False,
        requirement_file_path=None,
    ):
        if custom_paths is None:
            custom_paths = set()

        if not excluded_paths:
            excluded_paths = set()

        if extra_apt_packages is None:
            extra_apt_packages = set()

        self.custom_paths = custom_paths
        self.requirements_file_path = requirements_file_path
        self.odoo_base_path = odoo_base_path
        self.excluded_paths = excluded_paths
        self.exclude_odoo = exclude_odoo
        self.disabled_modules = disabled_modules
        self.odoo_rc = odoo_rc or self.default_odoorc()
        self.include_odoo_entrypoints = include_odoo_entrypoints
        self.force_addons_lookup = force_addons_lookup
        self.init_logger = init_logger
        self.run_only = run_only
        self.extra_apt_packages = extra_apt_packages
        self.apt_install_recommends = apt_install_recommends
        self.package_map_file = package_map_file
        self.skip_pip = skip_pip
        self.skip_sudo_entrypoint = skip_sudo_entrypoint
        self.skip_postgres_wait = skip_postgres_wait
        self.allow_dangerous_settings = allow_dangerous_settings
        self.odoo_version = odoo_version
        self.master_password = master_password
        self.show_master_password = show_master_password
        self.strict_mode = strict_mode
        self.reset_access_rights = reset_access_rights
        self.requirement_file_path = requirement_file_path

    def default_odoorc(self):
        directories = [
            Path.cwd(),
        ]

        if 'HOME' in os.environ:
            path = Path(os.environ['HOME'])
            if path not in directories:
                directories.append(path)

        var_lib_path = Path("/var/lib/odoo")
        if var_lib_path not in directories:
            directories.append(var_lib_path)

        filenames = [
            '.odoorc',
            '.openerp_serverrc',
            'odoo.cfg'
        ]

        for directory in directories:
            for filename in filenames:
                rc_file = directory / filename
                if rc_file.exists():
                    return rc_file
        else:
            return Path.cwd() / 'odoo.cfg'

    @classmethod
    def from_env(klass, envvars=None):
        """
        Creates a ``Context`` from environment variables.
        """
        if envvars is None:
            envvars = EnvironmentVariables()

        args = {
            "custom_paths": set()
        }

        if envvars.ODOO_EXTRA_PATHS:
            args['custom_paths'] |= {
                Path(path)
                for path in envvars.ODOO_EXTRA_PATHS
            }

        if envvars.ODOO_STRICT_MODE is not None:
            args['strict_mode'] = envvars.ODOO_STRICT_MODE

        if envvars.ODOO_REQUIREMENTS_FILE:
            args['requirements_file_path'] = Path(
                envvars.ODOO_REQUIREMENTS_FILE
            )
        else:
            args['requirements_file_path'] = Path(
                '/var/lib/odoo/requirements.txt'
            )

        if envvars.ODOO_BASE_PATH:
            args['odoo_base_path'] = envvars.ODOO_BASE_PATH

        if envvars.ODOO_EXCLUDED_PATHS:
            args['excluded_paths'] = {
                Path(path)
                for path in envvars.ODOO_EXCLUDED_PATHS
            }

        if envvars.ODOO_DISABLED_MODULES:
            args['disabled_modules'] = envvars.ODOO_DISABLED_MODULES

        if envvars.ODOO_RC:
            args['odoo_rc'] = Path(envvars.ODOO_RC)

        args['init_logger'] = envvars.USE_ODOO_LOGGER

        args['extra_apt_packages'] = envvars.ODOO_EXTRA_APT_PACKAGES
        args['apt_install_recommends'] = envvars.APT_INSTALL_RECOMMENDS
        args['package_map_file'] = envvars.PACKAGE_MAP_FILE

        # Entrypoint
        args['skip_sudo_entrypoint'] = envvars.SKIP_SUDO_ENTRYPOINT
        args['skip_pip'] = envvars.SKIP_PIP
        args['skip_postgres_wait'] = envvars.SKIP_POSTGRES_WAIT
        args['allow_dangerous_settings'] = envvars.ALLOW_DANGEROUS_SETTINGS
        args['odoo_version'] = envvars.ODOO_VERSION
        args['master_password'] = envvars.MASTER_PASSWORD
        args['show_master_password'] = envvars.SHOW_MASTER_PASSWORD
        args['reset_access_rights'] = envvars.RESET_ACCESS_RIGHTS
        args['requirement_file_path'] = (
            Path(envvars.REQUIREMENTS_FILE_PATH)
            if envvars.REQUIREMENTS_FILE_PATH
            else None
        )

        return Context(**args)



from odoo_tools.utilities.requirements import merge_requirements


def test_merge_requirements(tmp_path):
    req1 = """cryptography
Pillow >3.0
invalid_module; python_version > '40.8'
"""
    req2 = """
gevent < 11
gevent > 2
Pillow <10
requests
requests[ssl] <3; python_version <= '4'
"""

    req3 = """git+https://github.com/path/to/package-two@41b95ec#egg=package-two
"""

    requirements = set()

    for index, req in enumerate([req1, req2, req3]):
        filename = tmp_path / f'requirements{index}.txt'
        with filename.open('w') as fout:
            fout.write(req)

        requirements.add(filename)

    result = merge_requirements(requirements)

    assert set(result) == set([
        'git+https://github.com/path/to/package-two@41b95ec#egg=package-two',
        'cryptography',
        'pillow <10,>3.0',
        'gevent <11,>2',
        'requests [ssl] <3'
    ])



import pytest
from odoo_tools.modules.search import Manifest
from odoo_tools.odoo import Environment


def test_packages(tmp_path):
    addons = tmp_path / 'addons'
    mod1 = Manifest(addons / 'mod1')
    mod1.set_attribute(['external_dependencies', 'python'], ['gitpython'])
    mod1.save()

    mod2 = Manifest(addons / 'mod2')
    mod2.set_attribute(['external_dependencies', 'python'], ['barcode'])
    mod2.save()

    env = Environment()
    env.context.custom_paths.add(addons)
    env.context.extra_apt_packages = {"python3"}

    package_file = addons / 'apt-packages.txt'

    with package_file.open('w') as fout:
        fout.write("""git
vim""")

    packages = env.manage.packages()

    assert packages == {"git", "vim", "python3"}


def test_management_config():
    env = Environment()
    with pytest.raises(SystemError):
        env.manage.config()



from collections import defaultdict

try:
    from pip._internal.network.session import PipSession
    from pip._internal.req.req_file import parse_requirements
    from pip._internal.req.constructors import (
        install_req_from_parsed_requirement
    )
except ImportError:
    from pip.download import PipSession
    from pip.req.req_file import parse_requirements

    def install_req_from_parsed_requirement(req):
        return req


class Requirement(object):
    def __init__(self):
        self.extras = set()
        self.specifiers = set()
        self.links = set()
        self.editable = False


def merge_requirements(files):
    requirements = defaultdict(lambda: Requirement())
    links = set()

    session = PipSession()

    for filename in files:
        filename_str = str(filename)
        f_requirements = parse_requirements(filename_str, session=session)
        for parsed_requirement in f_requirements:
            requirement = install_req_from_parsed_requirement(
                parsed_requirement
            )

            if requirement.markers and not requirement.markers.evaluate():
                continue

            if not hasattr(requirement.req, 'name'):
                links.add(requirement.link.url)
                break
            name = requirement.req.name.lower()
            specifiers = requirement.req.specifier
            extras = requirement.req.extras
            requirements[name].extras |= set(extras)
            requirements[name].specifiers |= set(specifiers)
            if requirement.link:
                requirements[name].links |= {requirement.link.url}
            requirements[name].editable |= requirement.editable

    result = []
    for key, value in requirements.items():
        if value.links:
            result.append("%s" % value.links.pop())
        else:
            requirement_line = [key]

            if value.extras:
                extras = [str(extra) for extra in value.extras]
                extras.sort()
                requirement_line.append("[{}]".format(
                    ",".join(extras)
                ))

            if value.specifiers:
                specifiers = [str(spec) for spec in value.specifiers]
                specifiers.sort()
                requirement_line.append(",".join(specifiers))

            result.append(" ".join(requirement_line))

    for link in links:
        result.append(link)

    return result



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="odoo-tools",
    version="0.1.1",
    author="Loïc Faure-Lacroix <lamerstar@gmail.com>",
    author_email="lamerstar@gmail.com",
    description="Odoo Tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://odoo-tools.readthedocs.io",
    project_urls={
        "Source": "https://github.com/llacroix/odoo-tools",
        "Documentation": "https://odoo-tools.readthedocs.io",
    },
    packages=setuptools.find_packages(),
    install_requires=[
        "giturlparse",
        "toposort",
        "toml",
        "requests",
        "packaging",
        "lxml",
        "docutils",
        "polib",
        "six>=1.12.0",
        "cryptography",
        "click",
        "passlib",
        "ptpython",
    ],
    extras_require={
        "docs": [
            "sphinx",
            "furo",
            "sphinx-argparse",
            "sphinx-click",
        ],
        "test": [
            "mock",
            "pytest",
            "pytest-cov"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points={
        "console_scripts": [
            "odootools = odoo_tools.cli.odot:command"
        ],
    },
    package_data={
        "odoo_tools": [
            "requirements/*.txt",
            "packages/*.toml",
        ],
    }
)



import click
from ...compat import Path


def path_complete(ctx, param, incomplete):
    current_path = Path(incomplete)

    if current_path.exists():
        base_path = current_path
    else:
        base_path = current_path.parent

    paths = [str(base_path)]
    for path in base_path.iterdir():
        if path.is_dir():
            paths.append(str(path))

    return paths


class ModuleType(click.ParamType):
    name = "module"

    def convert(self, value, param, ctx):
        if isinstance(value, set):
            return value

        return {
            val.strip()
            for val in value.split(',')
            if val.strip()
        }


MODULE_TYPE = ModuleType()



import tempfile
from collections import defaultdict
from ..compat import Path
from ..modules.search import find_modules_paths
from ..utilities.requirements import merge_requirements


class ModuleApi(object):
    def __init__(self, environment):
        self.environment = environment
        self._all_manifests = None
        self._manifest_by_name = None

    def list(self, reload=False, filters=None):
        if filters is None:
            filters = set(['installable'])

        if not self._all_manifests or reload:
            all_manifests = find_modules_paths(
                self.environment.addons_paths(),
                filters,
                self.environment.context
            )
            self._all_manifests = all_manifests

        return self._all_manifests

    def get(self, name):
        if not self._manifest_by_name:
            self._manifest_by_name = defaultdict(list)
            for module in self.list():
                self._manifest_by_name[module.technical_name].append(module)

        mods = self._manifest_by_name[name]

        return mods[0]

    def server_wide_modules(self):
        """
        Search in the modules available in the environment for modules
        that are marked with the non standard property ``server_wide``.

        If a manifest is found with ``server_wide`` set to True. It will
        be returned as a server wide module.

        By default, it will always return the module ``base`` and ``web``
        as server wide modules.

        Example of use:

        .. code:: python

            with env.config():
                env.set_config(
                    'server_wide_modules',
                    ",".join(env.server_wide_modules())
                )

        Returns:
            modules (List<str>): List of server wide module names.

        """
        base_server_wide_modules = ['base', 'web']

        custom_server_wide_modules = [
            manifest.technical_name
            for manifest in self.list()
            if manifest.server_wide
        ]

        return base_server_wide_modules + custom_server_wide_modules

    def disabled_modules(self):
        """
        Generator returning a list of disabled modules based on the
        ODOO_DISABLED_MODULES environment variable.

        Example:

        .. code:: python

            for manifest in env.disabled_modules():
                manifest.remove()

        Returns:
            modules (List<Manifest>): A list of Manifest object that
                represent the found modules.
        """

        if not self.environment.context.disabled_modules:
            return

        all_modules = self.list()
        for module in all_modules:
            if module.name in self.environment.context.disabled_modules:
                yield module

    def remove_disabled(self):
        for module in self.disabled_modules():
            module.remove()

    def requirements(
        self,
        lookup_requirements=False,
        package_map=None,
        extra_paths=None,
        extra_rules=None
    ):
        if package_map is None:
            package_map = {}

        if extra_paths is None:
            extra_paths = set()

        if extra_rules is None:
            extra_rules = set()

        filters = set(['installable', 'python_dependencies'])
        modules = self.list(filters=filters)

        packages = set()
        for module in modules:
            packages |= module.requirements(package_map=package_map)

        if extra_rules or packages:
            name_template = "{}.txt"
            temp_file_count = 0

            with tempfile.TemporaryDirectory() as tempdir:
                folder_root = Path(tempdir)

                for rule in set(extra_rules) | packages:
                    file_name = name_template.format(temp_file_count)
                    file_handle = folder_root / file_name

                    with file_handle.open("wb") as temp:
                        temp.write(rule.encode('utf-8'))

                    extra_paths.add(str(file_handle))

                    temp_file_count += 1

                requirements = merge_requirements(extra_paths)
        else:
            requirements = merge_requirements(extra_paths)

        return set(requirements)



import giturlparse

from . import fields
from . import models


class ValidationError(Exception):
    pass


class Version(object):

    def __init__(self, since, until=None):
        self.since = since
        self.until = until


class KeyValue(models.Extendable, models.ParentedObject, models.BaseModel):
    def extend(self, other):
        options = self._data.copy()
        options.update(other._data if other and other._data else {})
        return KeyValue(options)

    def to_dict(self):
        return self._data.copy()


class RepoConfig(models.ParentedObject, models.Extendable, models.BaseModel):

    #: (str): Url of the repository. The url can point
    #: to a repository using the ssh format or https format.
    url = fields.Url(
        "Url of repository",
        availability=Version(since=1),
        default=None,
    )

    #: (str): the commit id to use
    commit = fields.String(
        "Commit of the object",
        availability=Version(since=1),
        default=None,
    )

    #: (str): the branch name to use
    branch = fields.String(
        "Branch of the object",
        availability=Version(since=1),
        default=None,
    )

    #: A private key if provided, the private key can
    #: be raw or it can be encrypted using Fernet encryption.
    private_key = fields.String(
        default=None,
    )

    #: Tell if the repo require authentication. In some cases,
    #: you could have https repositories that require extra
    #: credentials. If auth is False, then the url will not
    #: get altered to use credentials like access token provided
    #: in a credentials store.
    auth = fields.Boolean(
        default=None
    )

    @property
    def ref(self):
        """
        Returns the proper ref to use.

        By default it will try to use the following values in that
        order:

        - commit
        - branch
        - resolved odoo version or None

        Returns:
            str | None: the default ref to use.
        """
        service = self.get_parent(ServiceManifest)

        if (
            service and
            service.resolved.odoo and
            service.resolved.odoo.version
        ):
            default = service.resolved.odoo.version
        else:
            default = None

        return self.commit or self.branch or default

    @property
    def repo_path(self):
        """
        Converts the url of the repo in a unique path.

        The main reason is to provide a path that can be used as
        a unique identifier for the repositories. When inheriting
        services from an other manifests, all projects related to
        the same url will be inherited accordingly.

        When fetching repositories, it ensure that a project a/web
        and b/web will not be fetched into a web folder. Or to some
        extent, a github.com/a/web and gitlab.com/a/web are still
        considered as two different projects.

        Returns:
            str: the path of the repo
        """
        url = giturlparse.parse(self.url.lower())

        if url.valid:
            path = "_".join([
                url.host.replace('.', '_'),
                url.owner.replace('/', '_'),
                url.repo.replace('/', '_')
            ])
        else:
            path = self.url

        return path

    def extend(self, other):
        if not other:
            return self

        return RepoConfig({
            "url": other.url or self.url,
            "commit": other.commit or self.commit,
            "branch": other.branch or self.branch,
            "private_key": other.private_key or self.private_key,
        })

    def to_dict(self):
        return {
            "url": self.url,
            "commit": self.commit,
            "branch": self.branch,
            "ref": self.ref,
            "private_key": self.private_key
        }


class OdooConfig(models.ParentedObject, models.Extendable, models.BaseModel):
    version = fields.String()

    repo = fields.Object(
        "Repository Configuration",
        object_class=RepoConfig,
        availability=Version(since=1)
    )

    options = fields.Object(
        "Odoo Custom Options",
        object_class=KeyValue,
        availability=Version(since=1)
    )

    def extend(self, other):
        if not other:
            return self

        return OdooConfig({
            "version": (
                other.version
                if other.version
                else self.version
            ),
            "repo": (
                other.repo.extend(self.repo)
                if other.repo
                else self.repo
            ),
            "options": (
                other.options.extend(self.options)
                if other.options
                else self.options
            )
        })

    def to_dict(self):
        return {
            "version": self.version,
            "repo": self.repo.to_dict() if self.repo else {},
            "options": self.options.to_dict() if self.options else {}
        }


class ManifestProxy(models.ParentedObject, models.BaseModel):
    reference = fields.String()

    @classmethod
    def parse(klass, data):
        return ManifestProxy({"reference": data})


class ServiceManifest(
    models.ParentedObject,
    models.Extendable,
    models.BaseModel
):
    """
    This object represent all properties that can be set
    on a manifest for odoo services.
    """

    #: (String): A string representing the name of the
    #: service configuration. For example, you may want to define
    #: a service staging and production with different settings.
    name = fields.String(
        "name",
        availability=Version(since="1")
    )

    #: (ServiceManifest): A reference to an other service manifest.
    #: It's possible to reference manifests by name. In that case, a
    #: service will be able to merge its own properties with a parent
    #: manifest configuration.
    #:
    #: For example, you may want to have a production server using the
    #: production branch of certain repositories. While a staging
    #: environment tracking the staging branch. To make things simpler,
    #: you could have all addons defined in the staging environment. But
    #: you could also define custom branches for production and keep the
    #: rest as the staging environment is using.
    inherit = fields.ProxyObject(
        "Reference to parent manifest",
        getter="get_inherited_object",
        availability=Version(since="1")
    )

    #: (OdooConfig): A field referencing an odoo configuration.
    odoo = fields.Object(
        "Odoo Configuration",
        object_class=OdooConfig,
        availability=Version(since="1")
    )

    #: (List<RepoConfig>): A list of repository configuration.
    addons = fields.Dict(
        "List of addons",
        object_class=RepoConfig,
        key="repo_path",
    )

    #: (KeyValue): A key value store of labels.
    labels = fields.Object(
        "Odoo Custom Options",
        object_class=KeyValue,
        availability=Version(since=1)
    )

    #: (KeyValue): A key value store of environment variables.
    env = fields.Object(
        "Odoo Custom Options",
        object_class=KeyValue,
        availability=Version(since=1)
    )

    def get_inherited_object(self, key):
        """
        Returns:
            ServiceManifest: Related service manifest given by its name.
        """
        parent = self.get_parent(ServiceManifests)
        return parent.services[key]

    # TODO refactor to make auto extandable objects
    # Guess how they can be extended without specifying
    # properties.
    def resolved_prop(self, prop_name):
        if not self.inherit:
            return getattr(self, prop_name)

        other_val = self.inherit.resolved_prop(prop_name)
        self_val = getattr(self, prop_name)

        if not other_val:
            return self_val

        if not self_val:
            return other_val

        # TODO return Extendable Collections in field with list/dict
        if isinstance(other_val, models.Extendable):
            return other_val.extend(self_val)
        elif isinstance(other_val, dict):
            ret_val = other_val.copy()
            for key, value in self_val.items():
                if (
                    key in ret_val and
                    isinstance(ret_val[key], models.Extendable)
                ):
                    ret_val[key] = ret_val[key].extend(value)
                else:
                    ret_val[key] = value
            return ret_val
        elif isinstance(other_val, list):
            return other_val + ret_val
        else:
            print("whoa")

    @property
    def resolved(self):
        """
        This service manifest with all of its
        properties resolved against the inherit property.

        For example, if you had a service that inherits from an
        other one and each of them had different addons configured.
        It would let you combine all configurations together.

        The value of this property would be a new ServiceManifest
        that has all addons of self and of inherit (recursively).

        Returns:
            ServiceManifest: ServiceManifest that is an extension of inherit
        """
        return self.extend(self.inherit)

    def extend(self, other):
        """
        Creates a new ServiceManifest that extend the other one.

        Args:
            other (ServiceManifest): The other manifest to inherit from.

        Returns:
            (ServiceManifest): A new ServiceManifest
        """
        if not other:
            return self

        return ServiceManifest({
            "name": self.name,
            "odoo": self.resolved_prop('odoo'),
            "addons": self.resolved_prop('addons'),
            "labels": self.resolved_prop('labels'),
            "env": self.resolved_prop('env'),
        })

    def to_dict(self):
        """
        Returns a dict representing the data in this object.

        Returns:
            dict: Dict containing all properties of this object.
        """
        inherit = (
            self._data['inherit']['ref']
            if 'inherit' in self._data
            else None
        )

        addons = self.addons or {}

        vals = {
            "name": self.name,
            "addons": {
                key: val.to_dict()
                for key, val in addons.items()
            }
        }

        if inherit:
            vals['inherit'] = inherit

        for attr in ['odoo', 'env', 'labels']:
            val = getattr(self, attr)
            if val:
                vals[attr] = val.to_dict()

        return vals


class ServiceManifests(models.BaseModel):
    #: (dict): Dictionary of service manifest by their name
    services = fields.Dict(
        "List of services",
        object_class=ServiceManifest,
        key="name"
    )

    def to_dict(self):
        """
        Returns a dict representing the data in this object.

        Returns:
            dict: Dict containing all properties of this object.
        """
        return {
            "services": {
                key: val.to_dict()
                for key, val in self.services.items()
            }
        }



import os
import pytest
from mock import patch, MagicMock
from odoo_tools.compat import SIGSEGV
from pathlib import Path

from odoo_tools.compat import (
    flush_streams,
    log,
    module_path,
    pipe,
)


def test_pipe():
    kv = {
    }

    with patch('odoo_tools.compat.subprocess') as sbp,\
         patch.dict(os.environ, kv, clear=True):

        popen = MagicMock()
        popen.returncode = 0
        sbp.Popen.return_value = popen

        ret = pipe(['odoo'])

        assert popen.wait.call_count == 1
        assert ret == 0
        assert os.environ == {}


def test_pipe_sigsegv_no_crash():
    kv = {
    }

    with patch('odoo_tools.compat.subprocess') as sbp,\
         patch.dict(os.environ, kv, clear=True):

        popen = MagicMock()
        popen.returncode = -SIGSEGV
        sbp.Popen.return_value = popen

        ret = pipe(['odoo'])

        assert popen.wait.call_count == 1
        assert ret == -SIGSEGV
        assert os.environ == {}


def test_log():
    log("message")
    log("message", 1, 2)


def test_flush():
    with patch('odoo_tools.compat.sys') as sis:
        assert sis.stdout.flush.call_count == 0
        assert sis.stderr.flush.call_count == 0
        flush_streams()
        assert sis.stdout.flush.call_count == 1
        assert sis.stderr.flush.call_count == 1
        flush_streams()
        assert sis.stdout.flush.call_count == 2
        assert sis.stderr.flush.call_count == 2


def test_module_path():
    with patch('odoo_tools.compat.find_spec') as fp:
        mm = MagicMock()
        mm.__bool__ = lambda self: True
        mm.origin = None
        mm.submodule_search_locations = []

        fp.return_value = mm

        with pytest.raises(ModuleNotFoundError):
            module_path('zurgbug')

        assert module_path('zurgbug', raise_not_found=False) is None


def test_module_path1():
    with patch('odoo_tools.compat.find_spec') as fp:
        mm = MagicMock()
        mm.__bool__ = lambda self: True
        mm.origin = None
        mm.submodule_search_locations = ['/a']

        fp.return_value = mm

        assert module_path('zurgbug') == Path('/a')


def test_module_path2():
    with patch('odoo_tools.compat.find_spec') as fp:
        mm = MagicMock()
        mm.__bool__ = lambda self: True
        mm.origin = Path('/a/b/__init__.py')
        mm.submodule_search_locations = ['/a']

        fp.return_value = mm

        assert module_path('zurgbug') == Path('/a/b')



import os
import sys
import click
import logging

from ...docker.sudo_entrypoint import (
    install_apt_packages,
    fix_access_rights,
    remove_sudo,
)

from ...docker.user_entrypoint import (
    call_sudo_entrypoint,
    install_master_password,
    setup_env_config,
    setup_server_wide_modules,
    get_pg_environ,
    setup_addons_paths,
    install_python_dependencies,
    wait_postgresql,
    run_command
)

from ...entrypoints import execute_entrypoint


_logger = logging.getLogger(__name__)


@click.group()
def entrypoint():
    pass


@entrypoint.command()
@click.pass_context
def preprocess(ctx):
    env = ctx.obj['env']

    if not env.context.skip_sudo_entrypoint:
        ret = call_sudo_entrypoint()

    execute_entrypoint('odoo_tools.preprocess', env)

    sys.exit(ret)


@entrypoint.command()
@click.argument('params', nargs=-1)
@click.pass_context
def user(ctx, params):
    env = ctx.obj['env']

    # Install apt package first then python packages
    if not env.context.skip_sudo_entrypoint:
        ret = call_sudo_entrypoint()
    else:
        ret = 0

    if ret not in [0, None]:
        return ret

    with env.config():
        if not env.context.skip_pip:
            install_python_dependencies(env)

        install_master_password(env)
        environ, params = get_pg_environ(env, params)
        os.environ.update(environ)
        setup_addons_paths(env)
        setup_server_wide_modules(env)
        setup_env_config(env)

    if not env.context.skip_postgres_wait:
        wait_postgresql()

    ret = run_command(params)
    sys.exit(ret)


@entrypoint.command()
@click.argument('params', nargs=-1)
@click.pass_context
def sudo(ctx, params):
    if os.getuid() != 0:
        _logger.error("Cannot call sudo entrypoint as user")

    env = ctx.obj['env']

    try:
        install_apt_packages(env)
        fix_access_rights(env)
        env.modules.remove_disabled()
    except SystemError:
        return remove_sudo()
        if env.context.strict_mode:
            raise

    return remove_sudo()



"""
Module
------

Collection of API that can be used to simply manipulate
an odoo environment.

There are multiple use cases in which you'd want to use an api
to manipulate an odoo environment.

Here are a few notable example:

- Setting up a docker image
- Configuring a CI pipeline to run tests
"""
from .api.environment import Environment



from mock import MagicMock, patch
from odoo_tools.entrypoints import execute_entrypoint, entrypoint, custom_entrypoints


def test_entrypoint():

    @entrypoint("test")
    def super_entrypoint(manage):
        manage.called += 1

    ctx = MagicMock()
    ctx.called = 0
    execute_entrypoint("toast", ctx)
    assert ctx.called == 0

    ctx = MagicMock()
    ctx.called = 0
    execute_entrypoint("test", ctx)
    assert ctx.called == 1

    custom_entrypoints['test'] = []


def test_entrypoint_pkg():

    def mocked_iter_entrypoints():
        mock_entrypoint = MagicMock()

        def cb(ctx):
            ctx.called += 1

        mock_entrypoint.load.return_value = cb

        return [mock_entrypoint]

    with patch('pkg_resources.iter_entry_points') as mock_pkg:
        mock_pkg.return_value = mocked_iter_entrypoints()

        ctx = MagicMock()
        ctx.called = 0
        execute_entrypoint("test2", ctx)

        assert ctx.called == 1



import json
import giturlparse
import click
import toml
import logging

from ...services.objects import ServiceManifests
from ...compat import Path
from ...configuration.git import (
    fetch_addons,
    checkout_repo
)


_logger = logging.getLogger(__name__)


@click.group('service')
def service():
    pass


@service.command('show')
@click.option(
    '--url',
    help="Default url for self"
)
@click.argument(
    'service_file',
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
)
@click.argument('env')
@click.pass_context
def show(ctx, service_file, env, url):
    with Path(service_file).open('r') as fin:
        data = toml.load(fin)

    manifests = ServiceManifests.parse(data)

    service = manifests.services.get(env)
    resolved_service = service.resolved

    result = resolved_service.to_dict()

    if url and 'self' in result['addons']:
        self_addons = result['addons']['self']
        self_addons['url'] = url

    print(
        json.dumps(
            result,
            indent=2,
            sort_keys=True
        )
    )


@service.command('checkout')
@click.option(
    '--cache',
)
@click.option(
    '--decrypt-key',
)
@click.argument('service_file')
@click.argument('environment')
@click.argument('target')
@click.pass_context
def checkout(ctx, service_file, environment, target, cache, decrypt_key):
    with Path(service_file).open('r') as fin:
        data = toml.load(fin)

    manifests = ServiceManifests.parse(data)

    service = manifests.services.get(environment)
    resolved_service = service.resolved

    if cache:
        fetch_path = cache
    else:
        fetch_path = target

    results = []

    for key, addon in resolved_service.addons.items():
        if not giturlparse.parse(addon.url).valid:
            _logger.info(
                "Skipping addon %s as it has an invalid url.", addon.url
            )
            continue

        target_path = Path.cwd() / target / addon.repo_path

        path, info = fetch_addons(addon, fetch_path, decrypt_key=decrypt_key)

        if fetch_path != target_path:
            checkout_repo(path, target_path)

        results.append(info)



import sys
import click
import json
from toposort import toposort_flatten
from .utils import path_complete
from ...compat import Path
from ...modules.search import build_dependencies


@click.group()
def module():
    pass


@module.command("ls")
@click.option(
    '--only-name',
    default=False,
    help="Only display module name instead of path",
    is_flag=True
)
@click.option(
    '--csv',
    default=False,
    help="Only display module name instead of path",
    is_flag=True
)
@click.option(
    '--installable',
    help="Output only installable modules",
    is_flag=True,
    default=False
)
@click.option(
    '--non-installable',
    help="Output only non installable modules",
    is_flag=True,
    default=False
)
@click.option(
    '--without-version',
    help="Display only modules without versions",
    is_flag=True,
    default=False
)
@click.option(
    '-m',
    '--modules',
    help="Only return modules that matches the provided module names.",
    multiple=True
)
@click.option(
    '-p',
    '--path',
    help="Location in which to search",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    shell_complete=path_complete,
    multiple=True

)
@click.option(
    '--sorted',
    help="Sort module in alphabetical number",
    is_flag=True
)
@click.pass_context
def list_modules(
    ctx,
    only_name,
    csv,
    installable,
    non_installable,
    without_version,
    modules,
    path,
    sorted
):
    env = ctx.obj['env']

    if path:
        env.context.force_addons_lookup = True
        for _path in path:
            env.context.custom_paths.add(
                Path(_path)
            )

    filters = set()

    if installable:
        filters.add('installable')

    if non_installable:
        filters.add('non_installable')

    def check_module(mod):
        ret = True

        if without_version:
            ret = ret and 'version' in mod

        if modules:
            ret = ret and mod.path.name in modules

        return ret

    mods = [
        mod.technical_name if only_name else str(mod)
        for mod in env.modules.list(filters=filters)
        if check_module(mod)
    ]

    if sorted:
        mods.sort()

    if csv:
        print(",".join(mods))
    else:
        for mod in mods:
            print(mod)


@module.command("show")
@click.argument(
    'module'
)
@click.pass_context
def show_module(ctx, module):
    env = ctx.obj['env']

    mods = env.modules.list()

    for mod in mods:
        if mod.path.name == module:
            break
    else:
        print("Module {} not found".format(module))
        sys.exit(1)

    print(json.dumps(mod.values()))


@module.command("deps")
@click.option(
    '--only-name',
    default=False,
    help="Only display module name instead of path",
    is_flag=True
)
@click.option(
    '--csv',
    default=False,
    help="Only display module name instead of path",
    is_flag=True
)
@click.option(
    '--installable',
    help="Output only installable modules",
    is_flag=True,
    default=False
)
@click.option(
    '--non-installable',
    help="Output only non installable modules",
    is_flag=True,
    default=False
)
@click.option(
    '--without-version',
    help="Display only modules without versions",
    is_flag=True,
    default=False
)
@click.option(
    '-m',
    '--modules',
    help="Only return modules that matches the provided module names.",
    multiple=True
)
@click.option(
    '--csv-modules',
    help="Find all dependencies for modules input as csv",
    default=""
)
@click.option(
    '-p',
    '--path',
    help="Location in which to search",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    shell_complete=path_complete,
    multiple=True

)
@click.option(
    '--auto',
    help="Include auto installed modules",
    is_flag=True
)
@click.option(
    '--include-modules',
    help="Remove logs and warnings.",
    is_flag=True,
    default=False
)
@click.option(
    '--quiet',
    help="Remove logs and warnings.",
    is_flag=True,
    default=False
)
@click.pass_context
def show_dependencies(
    ctx,
    only_name,
    csv,
    installable,
    non_installable,
    without_version,
    modules,
    csv_modules,
    path,
    auto,
    quiet,
    include_modules
):
    env = ctx.obj['env']

    if path:
        env.context.force_addons_lookup = True
        for _path in path:
            env.context.custom_paths.add(
                Path(_path)
            )

    filters = set()

    if installable:
        filters.add('installable')

    if non_installable:
        filters.add('non_installable')

    def check_module(mod):
        ret = True

        if without_version:
            ret = ret and 'version' in mod

        return ret

    check_modules = list(modules) + [
        mod.strip()
        for mod in csv_modules.split(',')
        if mod.strip()
    ]

    modules_kv = {
        mod.path.name: mod
        for mod in env.modules.list(filters=filters)
        if check_module(mod)
    }

    dependencies = build_dependencies(
        modules_kv,
        check_modules[:],
        lookup_auto_install=auto,
        quiet=quiet
    )

    sorted_dependencies = []
    for dep in toposort_flatten(dependencies):
        if not include_modules and dep in modules:
            continue
        try:
            sorted_dependencies.append(modules_kv[dep])
        except KeyError:
            pass

    mods = [
        mod.path.name if only_name else str(mod)
        for mod in sorted_dependencies
    ]

    if csv:
        print(",".join(mods), end="")
    else:
        for mod in mods:
            print(mod)

@module.command(
    help="List requirements required by modules in addons_paths"
)
@click.option(
    '--exclude',
    help="Exclude file paths",
    multiple=True
)
@click.option(
    '-p',
    '--path',
    help="Add custom requirements file path",
    multiple=True
)
@click.option(
    '--add-rule',
    help="Add extra requirements rule",
    multiple=True
)
@click.option(
    '--package-map',
    help="Package map to resolve some modules that can't be installed",
)
@click.option(
    '--lookup-requirements',
    help=(
        "Lookup recursively requirements in addons_paths for a file "
        "named requirements.txt"
    ),
    is_flag=True,
    default=False
)
@click.option(
    '--sort',
    help="Sort requirements in alphabetical order",
    is_flag=True,
    default=False
)
@click.pass_context
def requirements(
    ctx,
    exclude,
    path,
    add_rule,
    package_map,
    lookup_requirements,
    sort
):
    env = ctx.obj['env']
    env.context.package_map_file = package_map

    found_files = env.requirement_files()

    for file_path in path:
        if file_path not in exclude:
            found_files.add(file_path)

    package_maps = env.package_map()

    requirements = env.modules.requirements(
        package_map=package_maps,
        extra_paths=found_files,
        extra_rules=set(add_rule)
    )

    if sort:
        requirements = [req for req in requirements]
        requirements.sort()

    print("\n".join(requirements))
